import pathlib
from collections.abc import Iterator

import pytest

from rbx.grading.caching import DependencyCache
from rbx.grading.judge.cacher import FileCacher
from rbx.grading.judge.sandbox import SandboxBase
from rbx.grading.judge.sandboxes.stupid_sandbox import StupidSandbox
from rbx.grading.judge.storage import FilesystemStorage, Storage


@pytest.fixture
def storage(request, cleandir: pathlib.Path) -> Iterator[Storage]:
    storage_path = cleandir / '.box' / '.storage'
    yield FilesystemStorage(storage_path)


@pytest.fixture
def file_cacher(request, storage: Storage) -> Iterator[FileCacher]:
    yield FileCacher(storage)


@pytest.fixture
def sandbox(request, file_cacher: FileCacher) -> Iterator[SandboxBase]:
    yield StupidSandbox(file_cacher=file_cacher)


@pytest.fixture
def dependency_cache(
    request, cleandir: pathlib.Path, storage: Storage
) -> Iterator[DependencyCache]:
    yield DependencyCache(cleandir / '.box', storage)



import sys

from rich.console import Console
from rich.theme import Theme

theme = Theme(
    {
        'default': 'bright_white',
        'rbx': 'bold italic yellow',
        'info': 'bright_black',
        'status': 'bright_white',
        'item': 'bold blue',
        'error': 'bold red',
        'success': 'bold green',
        'lnumber': 'dim cyan',
        'warning': 'bold yellow',
    }
)
console = Console(theme=theme, style='info', highlight=False)
stderr_console = Console(theme=theme, style='info', highlight=False, stderr=True)


def multiline_prompt(text: str) -> str:
    console.print(f'{text} (Ctrl-D to finish):\n')
    lines = sys.stdin.readlines()
    console.print()
    return ''.join(lines)



from typing import List, Optional

from rbx.grading import steps
from rbx.grading.caching import DependencyCache, NoCacheException
from rbx.grading.judge.sandbox import SandboxBase, SandboxParams
from rbx.grading.steps import (
    GradingArtifacts,
    GradingLogsHolder,
    RunLog,
    RunLogMetadata,
)


def compile(
    commands: List[str],
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
    dependency_cache: DependencyCache,
):
    ok = True
    with dependency_cache(
        commands, [artifacts], params.get_cacheable_params()
    ) as is_cached:
        if not is_cached and not steps.compile(
            commands=commands,
            params=params,
            artifacts=artifacts,
            sandbox=sandbox,
        ):
            ok = False
            raise NoCacheException()

    return ok


def run(
    command: str,
    params: SandboxParams,
    sandbox: SandboxBase,
    artifacts: GradingArtifacts,
    dependency_cache: DependencyCache,
    metadata: Optional[RunLogMetadata] = None,
) -> Optional[RunLog]:
    artifacts.logs = GradingLogsHolder()

    with dependency_cache(
        [command], [artifacts], params.get_cacheable_params()
    ) as is_cached:
        if not is_cached:
            steps.run(
                command=command,
                params=params,
                artifacts=artifacts,
                sandbox=sandbox,
                metadata=metadata,
            )

    return artifacts.logs.run



import pathlib
import shlex
from typing import Dict, List, Optional, Set, Tuple

import typer
from pydantic import BaseModel

from rbx import console
from rbx.box import package
from rbx.box.code import compile_item, run_item
from rbx.box.schema import CodeItem, Primitive
from rbx.box.testcases import find_built_testcase_inputs
from rbx.grading.judge.sandbox import SandboxBase
from rbx.grading.steps import (
    DigestHolder,
    DigestOrDest,
    DigestOrSource,
    GradingFileOutput,
)
from rbx.utils import StatusProgress

HitBounds = Dict[str, Tuple[bool, bool]]


class TestcaseValidationInfo(BaseModel):
    group: str
    path: pathlib.Path
    ok: bool
    hit_bounds: HitBounds
    message: Optional[str] = None


def _compile_validator(validator: CodeItem) -> str:
    try:
        digest = compile_item(validator)
    except:
        console.console.print(
            f'[error]Failed compiling validator [item]{validator.path}[/item][/error]'
        )
        raise
    return digest


def _bounds_or(lhs: Tuple[bool, bool], rhs: Tuple[bool, bool]) -> Tuple[bool, bool]:
    return (lhs[0] or rhs[0], lhs[1] or rhs[1])


def _process_bounds(log: str) -> HitBounds:
    bounds: HitBounds = {}
    for line in log.splitlines():
        items = line.split(':')
        if len(items) != 2:
            continue
        k, v = items
        v = v.strip()

        if 'constant-bounds' in k:
            continue

        hit = ('min-value-hit' in v, 'max-value-hit' in v)
        if k not in bounds:
            bounds[k] = hit
            continue
        bounds[k] = _bounds_or(bounds[k], hit)
    return bounds


def _merge_hit_bounds(hit_bounds: List[HitBounds]) -> HitBounds:
    res: HitBounds = {}
    for hb in hit_bounds:
        for k, hit in hb.items():
            if k not in res:
                res[k] = hit
                continue
            res[k] = _bounds_or(res[k], hit)
    return res


def _has_group_specific_validator() -> bool:
    pkg = package.find_problem_package_or_die()

    return any(group.validator is not None for group in pkg.testcases)


def _validate_testcase(
    testcase: pathlib.Path,
    validator: CodeItem,
    validator_digest: str,
    vars: Optional[Dict[str, Primitive]] = None,
) -> Tuple[bool, Optional[str], HitBounds]:
    vars = vars or {}
    for var in vars:
        assert (
            var.isidentifier()
        ), f'Variable {var} should be a valid Python identifier.'
    # TODO: check if needs to do some escaping
    var_args = [f'--{k}={v}' for k, v in vars.items()]
    var_args.extend(['--testOverviewLogFileName', 'validator.log'])

    message_digest = DigestHolder()
    log_digest = DigestHolder()
    run_log = run_item(
        validator,
        DigestOrSource.create(validator_digest),
        stdin=DigestOrSource.create(testcase),
        stderr=DigestOrDest.create(message_digest),
        outputs=[
            GradingFileOutput(
                src=pathlib.Path('validator.log'),
                digest=log_digest,
                optional=True,
            )
        ],
        extra_args=shlex.join(var_args) if var_args else None,
    )

    if (
        run_log is not None
        and run_log.exitcode != 0
        and run_log.exitstatus != SandboxBase.EXIT_NONZERO_RETURN
    ):
        console.console.print(
            f'[error]Validator [item]{validator.path}[/item] failed unexpectedly.[/error]'
        )
        console.console.print(f'[error]Summary:[/error] {run_log.get_summary()}')
        raise typer.Exit(1)

    log_overview = ''
    if log_digest.value is not None:
        log_overview = package.get_digest_as_string(log_digest.value or '')
    message = package.get_digest_as_string(message_digest.value or '')
    return (
        run_log is not None and run_log.exitcode == 0,
        message,
        _process_bounds(log_overview or ''),
    )


def validate_test(
    testcase: pathlib.Path,
    validator: CodeItem,
    validator_digest: str,
) -> Tuple[bool, Optional[str], HitBounds]:
    pkg = package.find_problem_package_or_die()
    return _validate_testcase(
        testcase, validator, validator_digest, vars=pkg.expanded_vars
    )


def compile_main_validator() -> Optional[Tuple[CodeItem, str]]:
    pkg = package.find_problem_package_or_die()
    if pkg.validator is None:
        return None

    return pkg.validator, _compile_validator(pkg.validator)


def validate_one_off(
    testcase: pathlib.Path,
    validator: CodeItem,
    validator_digest: str,
) -> TestcaseValidationInfo:
    ok, message, _ = validate_test(testcase, validator, validator_digest)
    info = TestcaseValidationInfo(
        group='interactive',
        path=testcase,
        ok=ok,
        hit_bounds={},
        message=message,
    )
    return info


def compile_validators(
    progress: Optional[StatusProgress] = None,
) -> Dict[str, str]:
    pkg = package.find_problem_package_or_die()

    group_to_compiled_digest = {}

    for group in pkg.testcases:
        validator = group.validator or pkg.validator
        if validator is None:
            continue
        if progress:
            progress.update(
                f'Compiling validator for group [item]{group.name}[/item]...'
            )
        group_to_compiled_digest[group.name] = _compile_validator(validator)

    return group_to_compiled_digest


def validate_testcases(
    progress: Optional[StatusProgress] = None,
    groups: Optional[Set[str]] = None,
) -> List[TestcaseValidationInfo]:
    def step():
        if progress is not None:
            progress.step()

    pkg = package.find_problem_package_or_die()

    group_to_compiled_digest = compile_validators(progress)

    validation_info = []

    for group in pkg.testcases:
        validator = group.validator or pkg.validator
        if validator is None:
            continue
        if group.name not in group_to_compiled_digest:
            continue
        if groups is not None and group.name not in groups:
            continue
        compiled_digest = group_to_compiled_digest[group.name]

        testcases = find_built_testcase_inputs(group)

        for testcase in testcases:
            ok, message, hit_bounds = validate_test(
                testcase, validator, compiled_digest
            )
            validation_info.append(
                TestcaseValidationInfo(
                    group=group.name,
                    path=testcase,
                    ok=ok,
                    hit_bounds=hit_bounds,
                    message=message,
                )
            )
            step()

    return validation_info


def has_validation_errors(infos: List[TestcaseValidationInfo]) -> bool:
    return any(not info.ok for info in infos)


def print_validation_report(infos: List[TestcaseValidationInfo]):
    console.console.rule('Validation report', style='status')
    hit_bounds_per_group: Dict[str, HitBounds] = {}
    for info in infos:
        if not info.ok:
            console.console.print(
                f'[error]Testcase [item]{info.path}[/item] failed verification:[/error]\n{info.message}'
            )
            continue

        if info.group not in hit_bounds_per_group:
            hit_bounds_per_group[info.group] = {}
        hit_bounds_per_group[info.group] = _merge_hit_bounds(
            [hit_bounds_per_group[info.group], info.hit_bounds]
        )

    if not hit_bounds_per_group:
        console.console.print()
        return

    if not _has_group_specific_validator():
        hit_bounds_per_group = {None: _merge_hit_bounds(hit_bounds_per_group.values())}

    for group, hit_bounds in hit_bounds_per_group.items():
        if group == 'samples':
            # Skip samples.
            continue
        if group is None:
            console.console.print('Hit bounds:')
        else:
            console.console.print(f'Group [item]{group}[/item] hit bounds:')

        for k, v in hit_bounds.items():
            if all(v):
                continue

            if not v[0]:
                console.console.print(
                    f'  - {k}: [warning]min-value not hit[/warning]',
                )
            if not v[1]:
                console.console.print(
                    f'  - {k}: [warning]max-value not hit[/warning]',
                )
        console.console.print()



import pathlib
import shutil
from typing import List

import typer
from pydantic import BaseModel

from rbx import console
from rbx.box import package
from rbx.box.package import get_build_testgroup_path, get_build_tests_path
from rbx.box.schema import Testcase, TestcaseGroup


class TestcaseData(BaseModel):
    input: str
    output: str


def find_built_testcases(group: TestcaseGroup) -> List[Testcase]:
    inputs = find_built_testcase_inputs(group)

    testcases = []
    for input in inputs:
        output = input.with_suffix('.out')
        testcases.append(Testcase(inputPath=input, outputPath=output))
    return testcases


def find_built_testcase_inputs(group: TestcaseGroup) -> List[pathlib.Path]:
    testgroup_path = get_build_testgroup_path(group.name)
    if not testgroup_path.is_dir():
        console.console.print(
            f'Testgroup {group.name} is not generated in build folder'
        )
        raise typer.Exit(1)

    return sorted(testgroup_path.glob('*.in'))


def clear_built_testcases():
    shutil.rmtree(str(get_build_tests_path()), ignore_errors=True)


def get_samples() -> List[Testcase]:
    tcs = find_built_testcases(package.get_testgroup('samples'))
    return [
        Testcase(
            inputPath=tc.inputPath.resolve(),
            outputPath=tc.outputPath.resolve()
            if tc.outputPath is not None and tc.outputPath.is_file()
            else None,
        )
        for tc in tcs
    ]



import pathlib
import shlex
import sys
from pathlib import PosixPath
from typing import List, Optional

import typer

from rbx.box import download, package
from rbx.box.environment import (
    ExecutionConfig,
    get_compilation_config,
    get_execution_config,
    get_extension_or_default,
    get_file_mapping,
    get_language,
    get_mapped_command,
    get_mapped_commands,
    get_sandbox_params_from_config,
    merge_execution_configs,
)
from rbx.box.extensions import MacExtension
from rbx.box.schema import CodeItem
from rbx.grading import steps_with_caching
from rbx.grading.steps import (
    DigestHolder,
    DigestOrDest,
    DigestOrSource,
    GradingArtifacts,
    GradingFileInput,
    GradingFileOutput,
    RunLog,
    RunLogMetadata,
)


def normalize_for_macos(commands: List[str]) -> List[str]:
    def normalize(command: str) -> str:
        extension = get_extension_or_default('mac', MacExtension)
        if extension.gpp_alternative is None:
            return command
        return command.replace('g++', extension.gpp_alternative)

    return [normalize(command) for command in commands]


def get_extension(code: CodeItem) -> str:
    path: pathlib.Path = PosixPath(code.path)
    return path.suffix[1:]


def find_language_name(code: CodeItem) -> str:
    if code.language is not None:
        return get_language(code.language).name
    return get_language(get_extension(code)).name


# Compile code item and return its digest in the storage.
def compile_item(code: CodeItem) -> str:
    generator_path = PosixPath(code.path)
    language = find_language_name(code)
    compilation_options = get_compilation_config(language)
    file_mapping = get_file_mapping(language)
    dependency_cache = package.get_dependency_cache()
    sandbox = package.get_singleton_sandbox()
    sandbox_params = get_sandbox_params_from_config(compilation_options.sandbox)

    if not compilation_options.commands:
        # Language is not compiled.
        return sandbox.file_cacher.put_file_from_path(generator_path)

    # Compile the generator
    commands = get_mapped_commands(compilation_options.commands, file_mapping)
    if sys.platform == 'darwin':
        commands = normalize_for_macos(commands)

    compiled_digest = DigestHolder()

    artifacts = GradingArtifacts()
    artifacts.inputs.extend(
        GradingFileInput(src=src, dest=dest)
        for src, dest in package.get_compilation_files(code)
    )
    download.maybe_add_testlib(code, artifacts)
    download.maybe_add_jngen(code, artifacts)
    artifacts.inputs.append(
        GradingFileInput(src=generator_path, dest=PosixPath(file_mapping.compilable))
    )
    artifacts.outputs.append(
        GradingFileOutput(
            src=PosixPath(file_mapping.executable),
            digest=compiled_digest,
            executable=True,
        )
    )

    if not steps_with_caching.compile(
        commands,
        params=sandbox_params,
        artifacts=artifacts,
        sandbox=sandbox,
        dependency_cache=dependency_cache,
    ):
        raise typer.Exit(1)

    assert compiled_digest.value is not None
    return compiled_digest.value


def run_item(
    code: CodeItem,
    executable: DigestOrSource,
    stdin: Optional[DigestOrSource] = None,
    stdout: Optional[DigestOrDest] = None,
    stderr: Optional[DigestOrDest] = None,
    inputs: Optional[List[GradingFileInput]] = None,
    outputs: Optional[List[GradingFileOutput]] = None,
    extra_args: Optional[str] = None,
    extra_config: Optional[ExecutionConfig] = None,
) -> Optional[RunLog]:
    language = find_language_name(code)
    execution_options = get_execution_config(language)
    if extra_config is not None:
        execution_options = merge_execution_configs([execution_options, extra_config])
    file_mapping = get_file_mapping(language)
    dependency_cache = package.get_dependency_cache()
    sandbox = package.get_singleton_sandbox()
    sandbox_params = get_sandbox_params_from_config(execution_options.sandbox)

    sandbox_params.set_stdall(
        stdin=PosixPath(file_mapping.input) if stdin is not None else None,
        stdout=PosixPath(file_mapping.output) if stdout is not None else None,
        stderr=PosixPath(file_mapping.error) if stderr is not None else None,
    )

    assert execution_options.command
    command = get_mapped_command(execution_options.command, file_mapping)

    if extra_args is not None:
        splitted_command = shlex.split(command)
        splitted_command.extend(shlex.split(extra_args))
        command = shlex.join(splitted_command)

    artifacts = GradingArtifacts()
    artifacts.inputs.append(
        GradingFileInput(
            **executable.expand(),
            dest=PosixPath(file_mapping.executable),
            executable=True,
        )
    )
    if stdin is not None:
        artifacts.inputs.append(
            GradingFileInput(
                **stdin.expand(),
                dest=PosixPath(file_mapping.input),
            )
        )
    if stdout is not None:
        artifacts.outputs.append(
            GradingFileOutput(
                src=PosixPath(file_mapping.output),
                **stdout.expand(),
            )
        )
    if stderr is not None:
        artifacts.outputs.append(
            GradingFileOutput(
                src=PosixPath(file_mapping.error),
                **stderr.expand(),
            )
        )
    if inputs:
        artifacts.inputs.extend(inputs)
    if outputs:
        artifacts.outputs.extend(outputs)

    return steps_with_caching.run(
        command,
        params=sandbox_params,
        sandbox=sandbox,
        artifacts=artifacts,
        dependency_cache=dependency_cache,
        metadata=RunLogMetadata(language=code.language),
    )



import atexit
import fcntl
import io
import logging
import os
import pathlib
import shutil
import tempfile
import typing
from typing import IO, List, Optional

import gevent

from rbx.grading.judge import digester, storage

logger = logging.getLogger(__name__)


class TombstoneError(RuntimeError):
    """An error that represents the file cacher trying to read
    files that have been deleted from the database.

    """

    pass


class FileCacher:
    """This class implement a local cache for files stored as FSObject
    in the database.

    """

    # This value is very arbitrary, and in this case we want it to be a
    # one-size-fits-all, since we use it for many conversions. It has
    # been chosen arbitrarily based on performance tests on my machine.
    # A few consideration on the value it could assume follow:
    # - The page size of large objects is LOBLKSIZE, which is BLCKSZ/4
    #   (BLCKSZ is the block size of the PostgreSQL database, which is
    #   set during pre-build configuration). BLCKSZ is by default 8192,
    #   therefore LOBLKSIZE is 2048. See:
    #   http://www.postgresql.org/docs/9.0/static/catalog-pg-largeobject.html
    # - The `io' module defines a DEFAULT_BUFFER_SIZE constant, whose
    #   value is 8192.
    # CHUNK_SIZE should be a multiple of these values.
    CHUNK_SIZE = 1024 * 1024  # 1 MiB

    backend: storage.Storage
    shared: bool
    file_dir: pathlib.Path
    temp_dir: pathlib.Path
    folder: Optional[pathlib.Path]

    def __init__(
        self,
        backend: storage.Storage,
        shared: bool = False,
        folder: Optional[pathlib.Path] = None,
    ):
        """Initialize."""

        self.backend = backend
        self.shared = shared
        self.folder = folder

        # First we create the config directories.
        if folder:
            self._create_directory_or_die(folder)

        if not self.is_shared():
            self.file_dir = pathlib.Path(tempfile.mkdtemp())
            # Delete this directory on exit since it has a random name and
            # won't be used again.
            atexit.register(lambda: shutil.rmtree(str(self.file_dir)))
        else:
            assert folder is not None
            self.file_dir = folder / 'fs-cache-shared'
        self._create_directory_or_die(self.file_dir)

        # Temp dir must be a subdirectory of file_dir to avoid cross-filesystem
        # moves.
        self.temp_dir = pathlib.Path(
            tempfile.mkdtemp(dir=self.file_dir, prefix='_temp')
        )
        atexit.register(lambda: shutil.rmtree(str(self.temp_dir)))
        # Just to make sure it was created.

    def is_shared(self) -> bool:
        """Return whether the cache directory is shared with other services."""
        return self.shared

    @staticmethod
    def _create_directory_or_die(directory: pathlib.Path):
        """Create directory and ensure it exists, or raise a RuntimeError."""
        directory.mkdir(parents=True, exist_ok=True)

    def precache_lock(self) -> Optional[IO[bytes]]:
        """Lock the (shared) cache for precaching if it is currently unlocked.

        Locking is optional: Any process can perform normal cache operations
        at any time whether the cache is locked or not.

        The locking mechanism's only purpose is to avoid wasting resources by
        ensuring that on each machine, only one worker precaches at any time.

        return (fileobj|None): The lock file if the cache was previously
            unlocked. Closing the file object will release the lock.
            None if the cache was already locked.

        """
        lock_file = self.file_dir / 'cache_lock'
        fobj = lock_file.open('wb')
        returned = False
        try:
            fcntl.flock(fobj, fcntl.LOCK_EX | fcntl.LOCK_NB)
        except BlockingIOError:
            # This exception is raised only if the errno is EWOULDBLOCK,
            # which means that the file is already locked.
            return None
        else:
            returned = True
            return fobj
        finally:
            if not returned:
                fobj.close()

    def _load(self, digest: str, cache_only: bool) -> Optional[IO[bytes]]:
        """Load a file into the cache and open it for reading.

        cache_only (bool): don't open the file for reading.

        return (fileobj): a readable binary file-like object from which
            to read the contents of the file (None if cache_only is True).

        raise (KeyError): if the file cannot be found.

        """
        cache_file_path = self.file_dir / digest

        if cache_only:
            if cache_file_path.exists():
                return None
        else:
            try:
                return cache_file_path.open('rb')
            except FileNotFoundError:
                pass

        logger.debug('File %s not in cache, downloading ' 'from database.', digest)

        ftmp_handle, temp_file_path = tempfile.mkstemp(dir=self.temp_dir, text=False)
        temp_file_path = pathlib.Path(temp_file_path)
        with open(ftmp_handle, 'wb') as ftmp, self.backend.get_file(digest) as fobj:
            storage.copyfileobj(fobj, ftmp, self.CHUNK_SIZE)

        if not cache_only:
            # We allow anyone to delete files from the cache directory
            # self.file_dir at any time. Hence, cache_file_path might no
            # longer exist an instant after we create it. Opening the
            # temporary file before renaming it circumvents this issue.
            # (Note that the temporary file may not be manually deleted!)
            fd = temp_file_path.open('rb')

        # Then move it to its real location (this operation is atomic
        # by POSIX requirement)
        temp_file_path.rename(cache_file_path)

        logger.debug('File %s downloaded.', digest)

        if not cache_only:
            return fd

    def cache_file(self, digest: str):
        """Load a file into the cache.

        Ask the backend to provide the file and store it in the cache for the
        benefit of future accesses, unless the file is already cached.
        Note that the cached file might still be deleted at any time, so it
        cannot be assumed to actually exist after this function completes.
        Always use the get_file* functions to access a file.

        digest (unicode): the digest of the file to get.

        raise (KeyError): if the file cannot be found.
        raise (TombstoneError): if the digest is the tombstone

        """
        if digest == storage.TOMBSTONE:
            raise TombstoneError()

        self._load(digest, True)

    def get_file(self, digest: str) -> IO[bytes]:
        """Retrieve a file from the storage.

        If it's available in the cache use that copy, without querying
        the backend. Otherwise ask the backend to provide it, and store
        it in the cache for the benefit of future accesses.

        The file is returned as a file-object. Other interfaces are
        available as `get_file_content', `get_file_to_fobj' and `get_
        file_to_path'.

        digest (unicode): the digest of the file to get.

        return (fileobj): a readable binary file-like object from which
            to read the contents of the file.

        raise (KeyError): if the file cannot be found.
        raise (TombstoneError): if the digest is the tombstone

        """
        if digest == storage.TOMBSTONE:
            raise TombstoneError()

        logger.debug('Getting file %s.', digest)

        return typing.cast(IO[bytes], self._load(digest, False))

    def path_for_symlink(self, digest: str) -> Optional[pathlib.Path]:
        if digest == storage.TOMBSTONE:
            raise TombstoneError()

        logger.debug('Getting symlink file path %s.', digest)
        return self.backend.path_for_symlink(digest)

    def get_file_content(self, digest: str) -> bytes:
        """Retrieve a file from the storage.

        See `get_file'. This method returns the content of the file, as
        a binary string.

        digest (unicode): the digest of the file to get.

        return (bytes): the content of the retrieved file.

        raise (KeyError): if the file cannot be found.
        raise (TombstoneError): if the digest is the tombstone

        """
        if digest == storage.TOMBSTONE:
            raise TombstoneError()
        with self.get_file(digest) as src:
            return src.read()

    def get_file_to_fobj(self, digest: str, dst: IO[bytes]):
        """Retrieve a file from the storage.

        See `get_file'. This method will write the content of the file
        to the given file-object.

        digest (unicode): the digest of the file to get.
        dst (fileobj): a writable binary file-like object on which to
            write the contents of the file.

        raise (KeyError): if the file cannot be found.
        raise (TombstoneError): if the digest is the tombstone

        """
        if digest == storage.TOMBSTONE:
            raise TombstoneError()
        with self.get_file(digest) as src:
            storage.copyfileobj(src, dst, self.CHUNK_SIZE)

    def get_file_to_path(self, digest: str, dst_path: pathlib.Path):
        """Retrieve a file from the storage.

        See `get_file'. This method will write the content of a file
        to the given file-system location.

        digest (unicode): the digest of the file to get.
        dst_path (string): an accessible location on the file-system on
            which to write the contents of the file.

        raise (KeyError): if the file cannot be found.

        """
        if digest == storage.TOMBSTONE:
            raise TombstoneError()
        with self.get_file(digest) as src:
            dst_path.parent.mkdir(parents=True, exist_ok=True)
            with dst_path.open('wb') as dst:
                storage.copyfileobj(src, dst, self.CHUNK_SIZE)

    def put_file_from_fobj(self, src: IO[bytes], desc: str = '') -> str:
        """Store a file in the storage.

        If it's already (for some reason...) in the cache send that
        copy to the backend. Otherwise store it in the file-system
        cache first.

        The file is obtained from a file-object. Other interfaces are
        available as `put_file_content', `put_file_from_path'.

        src (fileobj): a readable binary file-like object from which
            to read the contents of the file.
        desc (unicode): the (optional) description to associate to the
            file.

        return (unicode): the digest of the stored file.

        """
        logger.debug('Reading input file to store on the database.')

        # Unfortunately, we have to read the whole file-obj to compute
        # the digest but we take that chance to save it to a temporary
        # path so that we then just need to move it. Hoping that both
        # locations will be on the same filesystem, that should be way
        # faster than reading the whole file-obj again (as it could be
        # compressed or require network communication).
        # XXX We're *almost* reimplementing copyfileobj.
        with tempfile.NamedTemporaryFile(
            'wb', delete=False, dir=str(self.temp_dir)
        ) as dst:
            d = digester.Digester()
            buf = src.read(self.CHUNK_SIZE)
            while len(buf) > 0:
                d.update(buf)
                while len(buf) > 0:
                    written = dst.write(buf)
                    # Cooperative yield.
                    gevent.sleep(0)
                    if written is None:
                        break
                    buf = buf[written:]
                buf = src.read(self.CHUNK_SIZE)
            digest = d.digest()
            dst.flush()

            logger.debug('File has digest %s.', digest)

            cache_file_path = self.file_dir / digest

            # Store the file in the backend. We do that even if the file
            # was already in the cache
            # because there's a (small) chance that the file got removed
            # from the backend but somehow remained in the cache.
            # We read from the temporary file before moving it to
            # cache_file_path because the latter might be deleted before
            # we get a chance to open it.
            with open(dst.name, 'rb') as src:
                pending_file = self.backend.create_file(digest)
                if pending_file is not None:
                    storage.copyfileobj(src, pending_file.fd, self.CHUNK_SIZE)
                    self.backend.commit_file(pending_file, desc)

            os.rename(dst.name, cache_file_path)

        return digest

    def put_file_content(self, content: bytes, desc: str = '') -> str:
        """Store a file in the storage.

        See `put_file_from_fobj'. This method will read the content of
        the file from the given binary string.

        content (bytes): the content of the file to store.
        desc (unicode): the (optional) description to associate to the
            file.

        return (unicode): the digest of the stored file.

        """
        with io.BytesIO(content) as src:
            return self.put_file_from_fobj(src, desc)

    def put_file_text(self, text: str, desc: str = '') -> str:
        return self.put_file_content(text.encode('utf-8'), desc)

    def put_file_from_path(self, src_path: pathlib.Path, desc: str = '') -> str:
        """Store a file in the storage.

        See `put_file_from_fobj'. This method will read the content of
        the file from the given file-system location.

        src_path (Path): an accessible location on the file-system
            from which to read the contents of the file.
        desc (unicode): the (optional) description to associate to the
            file.

        return (unicode): the digest of the stored file.

        """
        with src_path.open('rb') as src:
            return self.put_file_from_fobj(src, desc)

    def describe(self, digest: str) -> str:
        """Return the description of a file given its digest.

        digest (unicode): the digest of the file to describe.

        return (unicode): the description of the file.

        raise (KeyError): if the file cannot be found.

        """
        if digest == storage.TOMBSTONE:
            raise TombstoneError()
        return self.backend.describe(digest)

    def get_size(self, digest: str) -> int:
        """Return the size of a file given its digest.

        digest (unicode): the digest of the file to calculate the size
            of.

        return (int): the size of the file, in bytes.

        raise (KeyError): if the file cannot be found.
        raise (TombstoneError): if the digest is the tombstone

        """
        if digest == storage.TOMBSTONE:
            raise TombstoneError()
        return self.backend.get_size(digest)

    def delete(self, digest: str):
        """Delete a file from the backend and the local cache.

        digest (unicode): the digest of the file to delete.

        """
        if digest == storage.TOMBSTONE:
            return
        self.drop(digest)
        self.backend.delete(digest)

    def drop(self, digest):
        """Delete a file only from the local cache.

        digest (unicode): the file to delete.

        """
        if digest == storage.TOMBSTONE:
            return
        cache_file_path: pathlib.Path = self.file_dir / digest
        cache_file_path.unlink(missing_ok=True)

    def purge_cache(self):
        """Empty the local cache.

        This function must not be called if the cache directory is shared.

        """
        self.destroy_cache()
        self.file_dir.mkdir(parents=True, exist_ok=True)
        if self.folder is not None:
            self.folder.mkdir(parents=True, exist_ok=True)

    def destroy_cache(self):
        """Completely remove and destroy the cache.

        Nothing that could have been created by this object will be
        left on disk. After that, this instance isn't usable anymore.

        This function must not be called if the cache directory is shared.

        """
        if self.is_shared():
            raise Exception('You may not destroy a shared cache.')
        shutil.rmtree(str(self.file_dir))

    def list(self) -> List[storage.FileWithDescription]:
        """List the files available in the storage.

        return ([(unicode, unicode)]): a list of pairs, each
            representing a file in the form (digest, description).

        """
        return self.backend.list()

    def check_backend_integrity(self, delete: bool = False) -> bool:
        """Check the integrity of the backend.

        Request all the files from the backend. For each of them the
        digest is recomputed and checked against the one recorded in
        the backend.

        If mismatches are found, they are reported with ERROR
        severity. The method returns False if at least a mismatch is
        found, True otherwise.

        delete (bool): if True, files with wrong digest are deleted.

        """
        clean = True
        for fwd in self.list():
            digest = fwd.filename
            d = digester.Digester()
            with self.backend.get_file(digest) as fobj:
                buf = fobj.read(self.CHUNK_SIZE)
                while len(buf) > 0:
                    d.update(buf)
                    buf = fobj.read(self.CHUNK_SIZE)
            computed_digest = d.digest()
            if digest != computed_digest:
                logger.error(
                    'File with hash %s actually has hash %s', digest, computed_digest
                )
                if delete:
                    self.delete(digest)
                clean = False

        return clean



import pathlib
from typing import List, Optional, Tuple

from rbx import config, hydration, metadata
from rbx.console import console
from rbx.schema import DumpedProblem, Testcase
from rbx.test import get_testcases_io


def get_testcase_paths(
    root: pathlib.Path, problem: DumpedProblem, i: int
) -> Tuple[pathlib.Path, pathlib.Path]:
    return (root / f'{problem.code}.{i}.in', root / f'{problem.code}.{i}.out')


def hydrate_problem(root: pathlib.Path, problem: DumpedProblem):
    for i, testcase in enumerate(problem.tests or []):
        in_path, out_path = get_testcase_paths(root, problem, i)
        in_path.write_text(testcase.input)
        out_path.write_text(testcase.output)


def add_testcase(root: pathlib.Path, problem: DumpedProblem, testcase: Testcase):
    problem_path = metadata.find_problem_path_by_code(problem.code, root)
    if not problem_path or not problem_path.is_file():
        console.print(
            f'[error]Problem [item]{problem.pretty_name()}[/item] not found.[/error]'
        )
        return

    # Pick next number.
    i = max([tc.index for tc in get_testcases_io(problem, root)] + [-1]) + 1
    in_path, out_path = get_testcase_paths(root, problem, i)
    in_path.write_text(testcase.input)
    out_path.write_text(testcase.output)

    console.print(
        f'Added testcase [item]{i}[/item] to problem [item]{problem.pretty_name()}[/item].'
    )


def remove_testcase(root: pathlib.Path, problem: DumpedProblem, i: int):
    problem_path = metadata.find_problem_path_by_code(problem.code, root)
    if not problem_path or not problem_path.is_file():
        console.print(
            f'[error]Problem [item]{problem.pretty_name()}[/item] not found.[/error]'
        )
        return

    testcases = get_testcases_io(problem, root)
    testcases = [testcase for testcase in testcases if testcase.index == i]
    if not testcases:
        console.print(
            f'[error]Testcase [item]{i}[/item] not found in problem [item]{problem.pretty_name()}[/item].[/error]'
        )
        return
    if testcases[0].input:
        testcases[0].input.unlink(missing_ok=True)
    if testcases[0].output:
        testcases[0].output.unlink(missing_ok=True)

    console.print(
        f'Removed testcase [item]{i}[/item] from problem [item]{problem.pretty_name()}[/item].'
    )


def edit_testcase(root: pathlib.Path, problem: DumpedProblem, i: int):
    problem_path = metadata.find_problem_path_by_code(problem.code, root)
    if not problem_path or not problem_path.is_file():
        console.print(
            f'[error]Problem [item]{problem.pretty_name()}[/item] not found.[/error]'
        )
        return

    testcases = get_testcases_io(problem, root)
    testcases = [testcase for testcase in testcases if testcase.index == i]
    if not testcases:
        console.print(
            f'[error]Testcase [item]{i}[/item] not found in problem [item]{problem.pretty_name()}[/item].[/error]'
        )
        return

    paths: List[Optional[pathlib.Path]] = [testcases[0].input, testcases[0].output]
    config.open_editor(*[path for path in paths if path is not None and path.is_file()])


def main(problem: Optional[str] = None):
    problems_to_hydrate = []
    if not problem:
        problems_to_hydrate = metadata.find_problems()
    else:
        dumped_problem = metadata.find_problem_by_anything(problem)
        problems_to_hydrate.append(dumped_problem)

    root = pathlib.Path()

    for dumped_problem in problems_to_hydrate:
        console.print(
            f'Hydrating problem [item]{dumped_problem.pretty_name()}[/item]...'
        )
        hydration.hydrate_problem(root, dumped_problem)



import typing

from pydantic import BaseModel

BocaLanguage = typing.Literal['c', 'cpp', 'cc', 'kt', 'java', 'py2', 'py3']

_MAX_REP_ERROR = 0.2  # 20% error allowed in time limit when adding reps


class BocaExtension(BaseModel):
    languages: typing.List[BocaLanguage] = list(typing.get_args(BocaLanguage))
    flags: typing.Dict[BocaLanguage, str] = {}
    maximumTimeError: float = _MAX_REP_ERROR

    def flags_with_defaults(self) -> typing.Dict[BocaLanguage, str]:
        res: typing.Dict[BocaLanguage, str] = {
            'c': '-std=gnu11 -O2 -lm -static',
            'cpp': '-O2 -lm -static',
            'cc': '-std=c++20 -O2 -lm -static',
        }
        res.update(self.flags)
        return res


class BocaLanguageExtension(BaseModel):
    # BocaLanguage this rbx language matches with.
    bocaLanguage: typing.Optional[str] = None



import pathlib
import tempfile
from typing import Type

import typer

from rbx import annotations, console
from rbx.box import builder, environment, package
from rbx.box.package import get_build_path
from rbx.box.packaging.boca.packager import BocaPackager
from rbx.box.packaging.packager import BasePackager, BuiltStatement
from rbx.box.packaging.polygon.packager import PolygonPackager
from rbx.box.statements.build_statements import build_statement

app = typer.Typer(no_args_is_help=True, cls=annotations.AliasGroup)


def run_packager(
    packager_cls: Type[BasePackager],
    verification: environment.VerificationParam,
) -> pathlib.Path:
    if not builder.verify(verification=verification):
        console.console.print(
            '[error]Build or verification failed, check the report.[/error]'
        )
        raise typer.Exit(1)

    pkg = package.find_problem_package_or_die()
    packager = packager_cls()

    statement_types = packager.statement_types()
    built_statements = []

    for statement_type in statement_types:
        languages = packager.languages()
        for language in languages:
            statement = packager.get_statement_for_language(language)
            statement_path = build_statement(statement, pkg, statement_type)
            built_statements.append(
                BuiltStatement(statement, statement_path, statement_type)
            )

    console.console.print(f'Packaging problem for [item]{packager.name()}[/item]...')

    with tempfile.TemporaryDirectory() as td:
        result_path = packager.package(
            get_build_path(), pathlib.Path(td), built_statements
        )

    console.console.print(
        f'[success]Problem packaged for [item]{packager.name()}[/item]![/success]'
    )
    console.console.print(f'Package was saved at [item]{result_path}[/item].')
    return result_path


@app.command('polygon', help='Build a package for Polygon.')
def polygon(
    verification: environment.VerificationParam,
):
    run_packager(PolygonPackager, verification=verification)


@app.command('boca', help='Build a package for BOCA.')
def boca(
    verification: environment.VerificationParam,
):
    run_packager(BocaPackager, verification=verification)



import pathlib

from rbx import annotations
from rbx.clone import create_problem_structure
from rbx.config import get_config
from rbx.console import console
from rbx.schema import Batch, Problem


def main(
    name: str,
    lang: annotations.Language,
    timelimit: annotations.Timelimit = 1000,
    memorylimit: annotations.Memorylimit = 256,
    multitest: annotations.Multitest = False,
):
    language = get_config().get_language(lang)
    if language is None:
        console.print(
            f'[error]Language {lang or get_config().defaultLanguage} not found in config. Please check your configuration.[/error]'
        )
        return

    problem = Problem(
        name=name,
        timeLimit=timelimit,
        memoryLimit=memorylimit,
        testType='multiNumber' if multitest else 'single',
        batch=Batch.create(),
    )
    create_problem_structure(
        pathlib.Path(),
        problem,
        language,
        status=None,
        verbose=True,
    )



import atexit
import pathlib

from rich.console import Console

from rbx import grading_utils
from rbx.grading.judge import cacher, storage
from rbx.grading.judge.sandboxes.isolate import IsolateSandbox

console = Console()


def main():
    fs = storage.FilesystemStorage(pathlib.PosixPath('/tmp/rbx-storage'))
    cache = cacher.FileCacher(fs)

    python_file = cache.put_file_text(
        """
#include <bits/stdc++.h>
 
int main() {
    std::cout << "Hello, World!" << std::endl;
    return 0;                                      
}
"""
    )

    sandbox = IsolateSandbox(
        cache, params=grading_utils.build_preprocess_sandbox_params(), debug=True
    )
    atexit.register(sandbox.cleanup)
    sandbox.create_file_from_storage(pathlib.PosixPath('run.cpp'), python_file)

    sandbox.params.stdout_file = pathlib.PosixPath('run.out')
    sandbox.params.stderr_file = pathlib.PosixPath('run.err')

    sandbox.execute_without_std(
        ['/usr/bin/g++', '-std=c++17', '-o', 'executable', 'run.cpp'],
    )
    try:
        sandbox.hydrate_logs()
    except Exception:
        console.print_exception()

    print(sandbox.log)
    print(sandbox.get_human_exit_description())
    print(sandbox.get_stats())

    print(sandbox.get_file_to_string(pathlib.PosixPath('run.out')))
    print(sandbox.get_file_to_string(pathlib.PosixPath('run.err')))


if __name__ == '__main__':
    main()



import atexit
import pathlib

from rich.console import Console

from rbx.grading.judge import cacher, storage
from rbx.grading.judge.sandboxes import stupid_sandbox

console = Console()


def main():
    fs = storage.FilesystemStorage(pathlib.PosixPath('/tmp/rbx-storage'))
    cache = cacher.FileCacher(fs)

    python_file = cache.put_file_text("print('hello')")

    sandbox = stupid_sandbox.StupidSandbox(cache)
    atexit.register(sandbox.cleanup)
    sandbox.create_file_from_storage(pathlib.PosixPath('run.py'), python_file)

    sandbox.params.stdout_file = pathlib.PosixPath('run.out')

    sandbox.execute_without_std(['ls'])
    try:
        sandbox.hydrate_logs()
    except Exception:
        console.print_exception()

    print(sandbox.get_human_exit_description())
    print(sandbox.get_stats())
    print(sandbox.log)

    print(sandbox.get_file_to_string(pathlib.PosixPath('run.out')))


if __name__ == '__main__':
    main()



import asyncio
import copy
import dataclasses
import fcntl
import os
import pty
import re
import signal
import struct
import termios
from typing import Callable, List, Optional

import pyte
import textual
from pyte.screens import Char
from rich.color import ColorParseError
from rich.segment import Segment
from rich.style import Style
from rich.text import Text
from textual import events
from textual.app import DEFAULT_COLORS
from textual.design import ColorSystem
from textual.geometry import Size
from textual.scroll_view import ScrollView
from textual.strip import Strip


class PyteDisplay:
    lines: List[List[Segment]]

    def __init__(self, lines):
        self.lines = lines

    @property
    def virtual_height(self):
        return len(self.lines)


@dataclasses.dataclass
class LogDisplayState:
    screen: pyte.Screen
    exitcode: Optional[int]


@dataclasses.dataclass
class Emulator:
    communicate_task: asyncio.Task
    send_task: asyncio.Task
    wait_task: asyncio.Task
    pid: int
    callback: Optional[Callable] = None

    def disconnect(self):
        self.communicate_task.cancel()
        self.send_task.cancel()
        self.wait_task.cancel()
        try:
            os.kill(self.pid, signal.SIGTERM)
        except OSError:
            # Process does not exist anymore.
            pass


class LogDisplay(ScrollView, can_focus=True):
    DEFAULT_CSS = """
    LogDisplay {
        background: $background;
    }
    """
    emulator: Optional[Emulator]
    exitcode: Optional[int]

    def __init__(
        self,
        default_colors: Optional[str] = 'textual',
        max_lines: int = 1000,
        name: Optional[str] = None,
        id: Optional[str] = None,
        classes: Optional[str] = None,
    ):
        super().__init__(name=name, id=id, classes=classes)

        self.emulator = None
        self.default_colors = default_colors
        if default_colors == 'textual':
            self.textual_colors = self.detect_textual_colors()

        self.virtual_size = Size(80, max_lines)
        self._max_lines = max_lines
        self._display = PyteDisplay([Text()])
        self._screen = pyte.Screen(self.virtual_size.width, self._max_lines)
        self.stream = pyte.Stream(self._screen)

        self.recv_queue = asyncio.Queue()
        self.send_queue = asyncio.Queue()
        self.exitcode = None

    async def on_resize(self, _event: events.Resize):
        if self.emulator is None:
            return

        # Update only width.
        self.virtual_size = Size(
            width=self.size.width - 2,  # Account for scroll bar.
            height=self.virtual_size.height,
        )
        await self.send_queue.put(
            ['set_size', self._max_lines, self.virtual_size.width]
        )
        self._screen.resize(self._max_lines, self.virtual_size.width)
        self.update_display()

    def update_display(self):
        lines: List[List[Segment]] = []
        for y in range(self._screen.lines):
            line = self._screen.buffer[y]
            line_segments = []
            accumulated_text = []
            for x in range(self._screen.columns):
                char: Char = line[x]
                if x > 0 and (not self.char_style_cmp(char, line[x - 1])):
                    text = ''.join(accumulated_text)
                    line_segments.append(
                        Segment(text, style=self.char_rich_style(line[x - 1]))
                    )
                    accumulated_text = []
                accumulated_text.append(char.data)
            if accumulated_text:
                text = ''.join(accumulated_text)
                line_segments.append(
                    Segment(
                        text, style=self.char_rich_style(line[self._screen.columns - 1])
                    )
                )
            lines.append(line_segments)

        # Remove empty lines from the back.
        while lines:
            last_line = lines[-1]
            text = ''.join(seg.text for seg in last_line)
            if text.strip():
                break
            lines.pop()

        self._display = PyteDisplay(lines)
        self.virtual_size = Size(
            width=self.size.width - 2,  # Account for possible vertical scrollbar.
            height=self._display.virtual_height,
        )
        self.refresh()

    def disconnect(self):
        if self.emulator is None:
            return
        cb = self.emulator.callback
        self.emulator.disconnect()
        self.emulator = None
        self.recv_queue = asyncio.Queue()
        self.send_queue = asyncio.Queue()
        if cb is not None:
            cb()
        self.recv_task.cancel()

    async def recv(self):
        while True:
            msg = await self.recv_queue.get()
            cmd = msg[0]
            if cmd == 'setup':
                await self.send_queue.put(
                    [
                        'set_size',
                        self.virtual_size.height,
                        self.virtual_size.width,
                    ]
                )
            elif cmd == 'stdout':
                chars = msg[1]
                self.stream.feed(chars)
                self.update_display()
            elif cmd == 'disconnect':
                self.disconnect()

    def char_rich_style(self, char: Char) -> Style:
        """Returns a rich.Style from the pyte.Char."""

        foreground = self.detect_color(char.fg)
        background = self.detect_color(char.bg)
        if self.default_colors == 'textual':
            if background == 'default':
                background = self.textual_colors['background']
            if foreground == 'default':
                foreground = self.textual_colors['foreground']

        style: Optional[Style]
        try:
            style = Style(
                color=foreground,
                bgcolor=background,
                bold=char.bold,
            )
        except ColorParseError as error:
            textual.log.warning('color parse error:', error)
            style = None

        return style or Style()

    def char_style_cmp(self, given: Char, other: Char) -> bool:
        """Compares two pyte.Chars and returns if these are the same.

        Returns:
            True    if char styles are the same
            False   if char styles differ
        """

        if (
            given.fg == other.fg
            and given.bg == other.bg
            and given.bold == other.bold
            and given.italics == other.italics
            and given.underscore == other.underscore
            and given.strikethrough == other.strikethrough
            and given.reverse == other.reverse
            and given.blink == other.blink
        ):
            return True

        return False

    def char_style_default(self, char: Char) -> bool:
        """Returns True if the given char has a default style."""

        if (
            char.fg == 'default'
            and char.bg == 'default'
            and char.bold is False
            and char.italics is False
            and char.underscore is False
            and char.strikethrough is False
            and char.reverse is False
            and char.blink is False
        ):
            return True

        return False

    def detect_color(self, color: str) -> str:
        """Tries to detect the correct Rich-Color based on a color name.

        * Returns #<color> if <color> is a hex-definition without "#"
        * Fixes wrong ANSI color names.

        Examples:
          * htop is using "brown" => not an ANSI color
        """

        if color == 'brown':
            return 'yellow'

        if color == 'brightblack':
            # fish tabbing through recommendations
            return '#808080'
        if color == 'brightwhite':
            return '#FFFFFF'

        if re.match('[0-9a-f]{6}', color, re.IGNORECASE):
            return f'#{color}'

        return color

    def detect_textual_colors(self) -> dict:
        """Returns the currently used colors of textual depending on dark-mode."""

        if self.app.dark:
            color_system: ColorSystem = DEFAULT_COLORS['dark']
        else:
            color_system: ColorSystem = DEFAULT_COLORS['light']

        return color_system.generate()

    def render_line(self, y: int) -> Strip:
        scroll_x, scroll_y = self.scroll_offset
        y += scroll_y
        if y >= len(self._display.lines):
            return Strip.blank(self.size.width)
        line = self._display.lines[y]
        strip = Strip(line).crop(scroll_x, scroll_x + self.size.width)
        return strip

    def export(self):
        return LogDisplayState(screen=copy.copy(self._screen), exitcode=self.exitcode)

    def load(self, state: LogDisplayState):
        self.disconnect()
        self._screen = state.screen
        self.stream = pyte.Stream(self._screen)
        self.exitcode = state.exitcode
        self.update_display()

    async def capture(self, argv: List[str]) -> int:
        self.exitcode = None
        self.recv_task = asyncio.create_task(self.recv())

        loop = asyncio.get_running_loop()
        send_queue = self.recv_queue
        recv_queue = self.send_queue
        event = asyncio.Event()

        pid, fd = pty.fork()
        if pid == 0:  # Child
            os.execvp(argv[0], argv)

        pout = os.fdopen(fd, 'w+b', 0)
        data: Optional[str] = None

        def on_output():
            nonlocal data
            try:
                data = pout.read(65536).decode()  # Read non-blocking.
            except Exception:
                data = None
                loop.remove_reader(pout)
            event.set()

        async def cleanup():
            try:
                loop.remove_reader(pout)
                _, exitstatus = os.waitpid(pid, os.WNOHANG)
                exitcode = os.waitstatus_to_exitcode(exitstatus)
                self.exitcode = exitcode
            except ChildProcessError:
                self.exitcode = -1
            await send_queue.put(['disconnect', 1])

        async def communicate():
            await send_queue.put(['setup', {}])

            while True:
                msg = await recv_queue.get()
                if msg[0] == 'set_size':
                    winsize = struct.pack('HH', msg[1], msg[2])
                    fcntl.ioctl(fd, termios.TIOCSWINSZ, winsize)

        async def send():
            while True:
                await event.wait()
                event.clear()
                if data is None:
                    await cleanup()
                else:
                    await send_queue.put(['stdout', data])

        async def wait():
            while True:
                try:
                    if os.waitpid(pid, os.WNOHANG) != (0, 0):
                        await cleanup()
                except ChildProcessError:
                    break
                await asyncio.sleep(0.5)

        finish = asyncio.Event()
        self.emulator = Emulator(
            asyncio.create_task(communicate()),
            asyncio.create_task(send()),
            asyncio.create_task(wait()),
            pid=pid,
            callback=lambda: finish.set(),
        )
        loop.add_reader(pout, on_output)
        await finish.wait()
        assert self.exitcode is not None
        return self.exitcode



from rbx.box import environment, package
from rbx.box.contest import contest_package


def clear_package_cache():
    pkgs = [package]

    for pkg in pkgs:
        for fn in pkg.__dict__.values():
            if hasattr(fn, 'cache_clear'):
                fn.cache_clear()


def clear_all_caches():
    pkgs = [package, environment, contest_package]

    for pkg in pkgs:
        for fn in pkg.__dict__.values():
            if hasattr(fn, 'cache_clear'):
                fn.cache_clear()



import functools
import pathlib
from typing import List, Optional

import typer
from pydantic import ValidationError

from rbx import console, utils
from rbx.box.contest.schema import Contest
from rbx.box.package import find_problem_package_or_die, warn_preset_deactivated
from rbx.box.schema import Package

YAML_NAME = 'contest.rbx.yml'


@functools.cache
def find_contest_yaml(root: pathlib.Path = pathlib.Path()) -> Optional[pathlib.Path]:
    root = root.resolve()
    contest_yaml_path = root / YAML_NAME
    while root != pathlib.PosixPath('/') and not contest_yaml_path.is_file():
        root = root.parent
        contest_yaml_path = root / YAML_NAME
    if not contest_yaml_path.is_file():
        return None
    warn_preset_deactivated(root)
    return contest_yaml_path


@functools.cache
def find_contest_package(root: pathlib.Path = pathlib.Path()) -> Optional[Contest]:
    contest_yaml_path = find_contest_yaml(root)
    if not contest_yaml_path:
        return None
    try:
        return utils.model_from_yaml(Contest, contest_yaml_path.read_text())
    except ValidationError as e:
        console.console.print(e)
        console.console.print('[error]Error parsing contest.rbx.yml.[/error]')
        raise typer.Exit(1) from e


def find_contest_package_or_die(root: pathlib.Path = pathlib.Path()) -> Contest:
    package = find_contest_package(root)
    if package is None:
        console.console.print(f'Contest not found in {root.absolute()}', style='error')
        raise typer.Exit(1)
    return package


def find_contest(root: pathlib.Path = pathlib.Path()) -> pathlib.Path:
    found = find_contest_yaml(root)
    if found is None:
        console.console.print(f'Contest not found in {root.absolute()}', style='error')
        raise typer.Exit(1)
    return found.parent


def within_contest(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        with utils.new_cd(find_contest()):
            return func(*args, **kwargs)

    return wrapper


def save_contest(
    package: Optional[Contest] = None, root: pathlib.Path = pathlib.Path()
) -> None:
    package = package or find_contest_package_or_die(root)
    contest_yaml_path = find_contest_yaml(root)
    if not contest_yaml_path:
        console.console.print(f'Contest not found in {root.absolute()}', style='error')
        raise typer.Exit(1)
    contest_yaml_path.write_text(utils.model_to_yaml(package))


def get_problems(contest: Contest) -> List[Package]:
    problems = []
    for problem in contest.problems:
        problems.append(find_problem_package_or_die(problem.get_path()))
    return problems
